const LOG = require("npm-niosz-log");
const log = LOG.newLog("oracle");
const oracledb = require("oracledb");
if (process.env.LD_LIBRARY_PATH) oracledb.initOracleClient({libDir: process.env.LD_LIBRARY_PATH});
const defaultConfigSize = 1000;
const uuid = require("uuid");
const fs = require("fs");
// ****************************************************************************
const generateFileName = function(i) {
    return i.toString().padStart(10,"0")+".json";
};
// ****************************************************************************
const oraCamelCase = function(column) {
    column = column.toLowerCase();
    column = column.split("_");
    for(var i=1;i<column.length;i++) {
        if (column[i].length>1) {
            column[i] = column[i][0].toUpperCase() + column[i].substring(1);
        } else {
            column[i] = column[i].toUpperCase();
        }
    }
    return column.join("");
};
// ****************************************************************************
const regex = /^[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}:[0-9]{2}\.[0-9]{3}Z$/gm;
const jsonDateParser = function(key,value) {
    if (typeof value === "string") {
        if (regex.exec(value)) 
            return new Date(value);
    }
    return value;
}
// ****************************************************************************
const createResultSet = function(folder,fetchs,rows,error) {
    return {
        folder          : folder,
        fetchs          : fetchs,
        remainingFetchs : fetchs,
        rows            : rows,
        remainingRows   : rows,
        error           : error,
        close           : function() { try{fs.rmSync(this.folder,{recursive:true,force:true});return true;}catch{return false;}; },
        fetch           : function(camelCaseOrMapper) {
            if (this.error) return false;
            try {
                var data = JSON.parse(fs.readFileSync(this.folder+"/"+generateFileName(this.fetchs-this.remainingFetchs),"utf8"),jsonDateParser);
                this.remainingFetchs--;
                this.remainingRows-=data.length;
                if (camelCaseOrMapper && data.length>0) {
                    if (typeof camelCaseOrMapper == "function") {
                        data = data.map(camelCaseOrMapper);
                    } else {
                        data = data.map((row) => {
                            var camelRow = {};
                            for (prop in row) {
                                const camelProp = oraCamelCase(prop);
                                camelRow[camelProp] = row[prop];
                            }
                            return camelRow;
                        });
                    }
                }
                return data;
            } catch {
                return false;
            }
        }
    };
};
// ****************************************************************************
const oraConnect = async function(connectionConfig) {
    try {
        var oracle = await oracledb.getConnection(connectionConfig);
        log.i("Connesso correttamente al database oracle");
        return oracle;
    } catch(e) {
        log.e("Connessione in errore : " + e);
        return null;
    }
};
// ****************************************************************************
const oraDisconnect = async function(connection,resultSet) {
    try {
        if (resultSet) try { await resultSet.resultSet.close(); } catch {};
        await connection.release();
        log.i("Disconnesso correttamente dal database Oracle");
    } catch(e) {
        log.w("Disconnesso con segnalazioni dal database Oracle:\n"+e);
    }
};
// ****************************************************************************
const oraSelect = async function(connectionConfig,queryConfig) {

    // *** connection to oracle
    var oracle = await oraConnect(connectionConfig);
    if (oracle == null) return null;

    // *** esecuzione query
    var queryData = null;
    try {
        queryData = await oracle.execute(queryConfig,[],
            {outFormat:oracledb.OUT_FORMAT_OBJECT,prefetchRows:defaultConfigSize,fetchArraySize:defaultConfigSize,resultSet:true});
        log.i("Query eseguita correttamente");
    } catch(e) {
        log.e("Errore nell'esecuzione della query : " + e);
        return null;
    }

    // *** prepare folder
    var folderData = null;
    try {
        folderData = "oradata/" + uuid.v4();
        fs.mkdirSync(folderData,{recursive:true});
        log.i("Creata cartella dati : " + folderData);
    } catch(e) {
        log.e("Errore nella creazione cartella dati : " + e);
        return null;
    }

    // *** fetching data
    var fetch = 0;
    var len = 0;
    var eof = false;
    while (!eof) {
        var rows = null;
        try {
            rows = await queryData.resultSet.getRows(defaultConfigSize);
        } catch(e) {
            log.e("Errore durante il fetching dei dati : " + e);
            await oraDisconnect(oracle,queryData);
            return createResultSet(folderData,fetch,len,true);
        }
        eof = (!rows.length) ? true : false;
        if (eof) continue;
        try {
            fs.writeFileSync(folderData+"/"+generateFileName(fetch),JSON.stringify(rows,null,2),"utf8");
        } catch(e) {
            log.e("Errore durante la scrittura fetching dati : " + e);
            await oraDisconnect(oracle,queryData);
            return createResultSet(folderData,fetch,len,true);
        }
        fetch++;
        len += rows.length;
        log.i("Record processati : " + len);
    }
    log.i("Record totali processati : " + len);

    // *** disconnection from oracle
    await oraDisconnect(oracle,queryData);
    
    // *** create management resultset response
    return createResultSet(folderData,fetch,len,false);

};
// ****************************************************************************
const ora = {
    select : oraSelect
};
// ****************************************************************************
module.exports = ora;